;;; -*- lexical-binding: t -*-
(defvar foo-bar nil
  "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
This is a multiline docstring where the first line is long.
foobar")
